
// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF MACHINE LEARNING RESEARCH (JMLR)
// AND JMLR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PUR- POSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JMLR, NOR JMLR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting
// the generality of the foregoing, neither the author, nor JMLR, nor
// JMLR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

// (C) 2011-2015 Urun Dogan, Tobias Glasmachers, Christian Igel.
// This program is part of the supplementary material of the JMLR
// paper 'A Unified View on Multi-class Support Vector Classification'.
// It requires version 3.0 of the Shark open source machine learning
// library. This program is free software; you can redistribute it
// and/or modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3, or
// (at your option) any later version.


#ifndef _sharedptr_H_
#define _sharedptr_H_


#include <cstdlib>


template <class T>
class SharedPtr
{
public:
	SharedPtr()
	: m_prev(this)
	, m_next(this)
	, m_data(NULL)
	{ }

	explicit SharedPtr(T* ptr)
	: m_prev(this)
	, m_next(this)
	, m_data(ptr)
	{ }

	SharedPtr(SharedPtr<T> const& other)
	: m_prev(this)
	, m_next(this)
	, m_data(const_cast<T*>(other.m_data))
	{
		if (m_data != NULL)
		{
			m_prev = const_cast<SharedPtr*>(&other);
			m_next = m_prev->m_next;
			m_prev->m_next = this;
			m_next->m_prev = this;
		}
	}

	~SharedPtr()
	{
		reset();
	}


	void operator = (T* ptr)
	{
		reset();
		m_data = ptr;
	}

	void operator = (SharedPtr<T> const& other)
	{
		reset();
		if (other.m_data != NULL)
		{
			m_data = const_cast<T*>(other.m_data);
			m_prev = const_cast<SharedPtr*>(&other);
			m_next = m_prev->m_next;
			m_prev->m_next = this;
			m_next->m_prev = this;
		}
	}

	void reset()
	{
		if (m_data != NULL)
		{
			if (m_prev == this)
			{
				delete m_data;
			}
			else
			{
				m_prev->m_next = m_next;
				m_next->m_prev = m_prev;
			}
			m_prev = this;
			m_next = this;
			m_data = NULL;
		}
	}

	T& operator * ()
	{ return *m_data; }

	T const & operator * () const
	{ return *m_data; }

	T* operator -> ()
	{ return m_data; }

	const T* operator -> () const
	{ return m_data; }

	bool operator == (const T* ptr) const
	{ return m_data == ptr; }
	bool operator == (SharedPtr const& ptr) const
	{ return m_data == ptr.m_data; }
	bool operator != (const T* ptr) const
	{ return m_data != ptr; }
	bool operator != (SharedPtr const& ptr) const
	{ return m_data != ptr.m_data; }
	bool operator < (const T* ptr) const
	{ return m_data < ptr; }
	bool operator < (SharedPtr const& ptr) const
	{ return m_data < ptr.m_data; }
	bool operator <= (const T* ptr) const
	{ return m_data <= ptr; }
	bool operator <= (SharedPtr const& ptr) const
	{ return m_data <= ptr.m_data; }
	bool operator > (const T* ptr) const
	{ return m_data > ptr; }
	bool operator > (SharedPtr const& ptr) const
	{ return m_data > ptr.m_data; }
	bool operator >= (const T* ptr) const
	{ return m_data >= ptr; }
	bool operator >= (SharedPtr const& ptr) const
	{ return m_data >= ptr.m_data; }

private:
	SharedPtr* m_prev;
	SharedPtr* m_next;
	T* m_data;
};


#endif
